class Key(object):
    '''
    A unique specification for routing data to a particular detector sub-system.
    At the core, a key is represented by 3-unsigned 1-byte integer fields which
    refer to an id code within a layer of the LArPix DAQ system heirarchy.
    Field 0 represents the io group id number, field 1 represents the io
    channel connecting to a MISO/MOSI pair, and field 2 represents the chip
    id. The io group is the device controlling a set of MOSI/MISO pairs, the
    io channel is a single MOSI/MISO pair controlling a collection of LArPix
    asics, and the chip id uniquely identifies a chip on a single MISO/MISO
    network.

    Each field should be a 1-byte unsigned integer (0-255) providing a unique
    lookup value for each component in the system. The id values of 0 and 255
    are reserved for special functionality.

    A key can be specified by a string of ``'<io group>-<io channel>-<chip id>'``, by io group, io channel, and chip id, or by
    using other Keys.

    Keys are hashed by their string representation and are equivalent to their
    string representation so::

        key = Key(1,1,1) # io group, io channel, chip id
        key == Key('1-1-1') # True
        key == Key(key) # True

        key == '1-1-1' # True
        key == (1,1,1) # True

        d = { key: 'example' }
        d[key] == 'example' # True
        d['1-1-1'] == 'example' # True

    Keys are "psuedo-immutable", i.e. you cannot change a Key's io_group,
    io_channel, or chip_id after it has been created.

    '''
    key_delimiter = '-'
    key_format = key_delimiter.join(('{io_group}', '{io_channel}', '{chip_id}'))

    def __init__(self, *args):
        self._initialized = False
        if len(args) == 3:
            self.io_group = args[0]
            self.io_channel = args[1]
            self.chip_id = args[2]
        elif len(args) == 1:
            if isinstance(args[0], Key):
                self.io_group = args[0].io_group
                self.io_channel = args[0].io_channel
                self.chip_id = args[0].chip_id
            elif isinstance(args[0], bytes):
                self.keystring = str(args[0].decode("utf-8"))
            else:
                self.keystring = str(args[0])
        else:
            raise TypeError('Key() takes 1 or 3 arguments ({} given)'.format(len(args)))
        self._initialized = True

    def __repr__(self):
        return 'Key(\'{}\')'.format(self.keystring)

    def __str__(self):
        return self.keystring

    def __eq__(self, other):
        if isinstance(other, Key):
            return self.io_group == other.io_group and \
            self.io_channel == other.io_channel \
            and self.chip_id == other.chip_id
        if isinstance(other, tuple):
            return self.io_group == other[0] and self.io_channel == other[1] \
            and self.chip_id == other[2]
        if str(self) == str(other):
            return True
        return False

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(str(self))

    def __getitem__(self, index):
        return (self.io_group, self.io_channel, self.chip_id)[index]

    @property
    def keystring(self):
        '''
        Key string specifying key io group, io channel, and chip id in the
        format: ``'<io group>-<io channel>-<chip id>'``
        '''
        return Key.key_format.format(
                io_group = self.io_group,
                io_channel = self.io_channel,
                chip_id = self.chip_id
            )

    @keystring.setter
    def keystring(self, val):
        if self._initialized:
            raise AttributeError('keystring cannot be modified')
        if not isinstance(val, str):
            raise TypeError('keystring must be str')
        parsed_keystring = val.split(Key.key_delimiter)
        if len(parsed_keystring) != 3:
            raise ValueError('invalid keystring formatting')
        self.io_group = int(parsed_keystring[0])
        self.io_channel = int(parsed_keystring[1])
        self.chip_id = int(parsed_keystring[2])

    @property
    def chip_id(self):
        '''
        1-byte unsigned integer representing the physical chip id (hardwired for
        v1 ASICs, assigned dynamically for v2 ASICs)
        '''
        return self._chip_id

    @chip_id.setter
    def chip_id(self, val):
        if self._initialized:
            raise AttributeError('chipid cannot be modified')
        chip_id = int(val)
        if chip_id > 255 or chip_id < 0:
            raise ValueError('chip_id must be 1-byte ({} invalid)'.format(chip_id))
        self._chip_id = chip_id

    @property
    def io_channel(self):
        '''
        1-byte unsigned integer representing the physical io channel. This
        identifies a single MOSI/MISO pair used to communicate with a single
        network of up to 254 chips.
        '''
        return self._io_channel

    @io_channel.setter
    def io_channel(self, val):
        if self._initialized:
            raise AttributeError('io_channel cannot be modified')
        io_channel = int(val)
        if io_channel > 255 or io_channel < 0:
            raise ValueError('io_channel must be 1-byte ({} invalid)'.format(io_channel))
        self._io_channel = io_channel

    @property
    def io_group(self):
        '''
        1-byte unsigned integer representing the physical device used to read
        out up to 254 io channels.
        '''
        return self._io_group

    @io_group.setter
    def io_group(self, val):
        if self._initialized:
            raise AttributeError('io_group cannot be modified')
        io_group = int(val)
        if io_group > 255 or io_group < 0:
            raise ValueError('io_group must be 1-byte ({} invalid)'.format(io_group))
        self._io_group = io_group

    @staticmethod
    def is_valid_keystring(keystring):
        '''
        Check if keystring can be interpreted as a larpix.Key

        :returns: ``True`` if the keystring can be interpreted as a larpix.Key
        '''
        if not isinstance(keystring, str):
            return False
        try:
            key = Key(keystring)
        except ValueError:
            return False
        return True

    def to_dict(self):
        '''
        Convert Key into a dict

        :returns: ``dict`` with ``'io_group'``, ``'io_channel'``, and ``'chip_id'``
        '''
        return_dict = dict(
                io_group = self.io_group,
                io_channel = self.io_channel,
                chip_id = self.chip_id
            )
        return return_dict

    @staticmethod
    def from_dict(d):
        '''
        Convert a dict into a Key object, dict must contain ``'io_group'``,
        ``'io_channel'``, and ``'chip_id'``

        :returns: ``Key``
        '''
        req_keys = ('io_group', 'io_channel', 'chip_id')
        if not all([key in d for key in req_keys]):
            raise ValueError('dict must specify {}'.format(req_keys))
        return Key(d['io_group'], d['io_channel'], d['chip_id'])



import pytest
import json
from larpix import configs

@pytest.fixture
def tmpfile(tmpdir):
    return str(tmpdir.join('test_conf.json'))

@pytest.fixture
def other_tmpfile(tmpdir):
    return str(tmpdir.join('other_test_conf.json'))

@pytest.fixture
def other_other_tmpfile(tmpdir):
    return str(tmpdir.join('other_other_test_conf.json'))

def write_json(file, **kwargs):
    with open(file,'w') as of:
        json.dump(kwargs, of)

def test_config_read(tmpfile):
    write_json(tmpfile,
        _config_type='test',
        test=True
        )

    config = configs.load(tmpfile, 'test')
    assert config['_config_type'] == 'test'
    assert config['test']

    config = configs.load(tmpfile)
    assert config['_config_type'] == 'test'
    assert config['test']

    with pytest.raises(AssertionError):
        config = configs.load(tmpfile, 'not_test')

def test_config_inheritance(tmpfile, other_tmpfile):
    write_json(tmpfile,
        _config_type='test',
        _include=[other_tmpfile],
        test=True
        )
    write_json(other_tmpfile,
        _config_type='test',
        inherited=True
        )

    config = configs.load(tmpfile, 'test')
    assert config['inherited']
    assert config['test']

    write_json(tmpfile,
        _config_type='test',
        _include=[other_tmpfile],
        test=True
        )
    write_json(other_tmpfile,
        _config_type='test',
        test=False
        )

    config = configs.load(tmpfile, 'test')
    assert config['test']

def test_config_inheritance_complex(tmpfile, other_tmpfile):
    write_json(tmpfile,
        _config_type='test',
        _include=[other_tmpfile],
        dict_field={'base':True,'dict':{'overwrite':True}},
        list_field=[1]
        )
    write_json(other_tmpfile,
        _config_type='test',
        dict_field={'inherited':True,'base':False,'dict':{'test':True,'overwrite':False}},
        list_field=[2]
        )
    config = configs.load(tmpfile)
    assert config['dict_field']['base']
    assert config['dict_field']['inherited']
    assert config['dict_field']['dict']['test']
    assert config['dict_field']['dict']['overwrite']
    assert config['list_field'] == [1]

def test_config_inheritance_order(tmpfile, other_tmpfile, other_other_tmpfile):
    write_json(tmpfile,
        _config_type='test',
        _include=[other_tmpfile, other_other_tmpfile]
        )
    write_json(other_tmpfile,
        _config_type='test',
        test=1
        )
    write_json(other_other_tmpfile,
        _config_type='test',
        test=2
        )

    config = configs.load(tmpfile)
    assert config['test'] == 2

def test_config_inheritance_error(tmpfile, other_tmpfile):
    write_json(tmpfile,
        _config_type='test',
        _include=[other_tmpfile]
        )
    write_json(other_tmpfile,
        _config_type='not_test'
        )

    with pytest.raises(AssertionError):
        config = configs.load(tmpfile, 'test')

    with pytest.raises(AssertionError):
        config = configs.load(tmpfile)



'''
The setup.py file for larpix-control.

'''

from setuptools import setup, find_packages
from codecs import open
import os

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(os.path.join(here, 'VERSION')) as f:
    version = f.read()

setup(
        name='larpix-control',
        version=version,
        description='Control the LArPix chip',
        long_description=long_description,
        long_description_content_type="text/markdown",
        url='https://github.com/larpix/larpix-control',
        author='Peter Madigan and Sam Kohn',
        author_email='pmadigan@lbl.gov',
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 2'
        ],
        keywords='dune physics',
        packages=find_packages(),
        install_requires=[
            'pyserial ~=3.4',
            'pytest ~=4.2',
            'bitarray ~=0.8',
            'pyzmq ~= 18.0',
            'sphinx_rtd_theme ~= 0.4.2',
            'numpy ~= 1.16',
            'h5py ~= 2.9',
            'bidict ~= 0.18.0',
            'networkx ~= 2.2'
            ],
        scripts=[
            'scripts/gen_controller_config.py',
            'scripts/gen_hydra_simple.py'
            ],
        package_data={
            'larpix.configs': ['*/*.json'],
            },
)



import time
import os

import numpy as np
import h5py

from larpix.logger import Logger
from larpix import Packet, TimestampPacket, Packet_v1, Packet_v2, SyncPacket, TriggerPacket
from larpix.format.hdf5format import to_file, latest_version

class HDF5Logger(Logger):
    '''
    The HDF5Logger is a logger class for logging packets to the LArPix+HDF5 format.

    The file format is implemented in ``larpix.format.hdf5format``,
    which also contains a function to convert back from LArPix+HDF5 to
    LArPix packets.

    :var data_desc_map: specifies the mapping between
        objects sent to the logger and the specific logger buffer to store them
        in. As of LArPix+HDF5 version 1.0 and larpix-control version
        2.3.0 there is only one buffer called ``'packets'`` which stores
        all of the data to send to LArPix+HDF5.

    :param filename: filename to store data (appended to ``directory``)
        (optional, default: ``None``)
    :param buffer_length: how many data messages to hang on to before flushing
        buffer to the file (optional, default: ``10000``)
    :param directory: the directory to save the data in (optional,
        default: '')
    :param version: the format version of LArPix+HDF5 to use (optional,
        default: ``larpix.format.hdf5format.latest_version``)

    '''
    data_desc_map = {
        Packet_v1: 'packets',
        Packet_v2: 'packets',
        TimestampPacket: 'packets',
        SyncPacket: 'packets',
        TriggerPacket: 'packets'
    }

    def __init__(self, filename=None, buffer_length=10000,
            directory='', version=latest_version, enabled=False):
        super(HDF5Logger, self).__init__(enabled=enabled)
        self.version = version
        self.filename = filename
        self.directory = directory
        self.datafile = None
        self.buffer_length = buffer_length

        self._buffer = {'packets': []}
        if not self.filename:
            self.filename = self._default_filename()
        self.filename = os.path.join(self.directory, self.filename)

    def _default_filename(self, timestamp=None):
        '''
        Fetch the default filename based on a timestamp

        :param timestamp: tuple or ``struct_time`` representing a timestamp
        '''
        log_prefix = 'datalog'
        time_format = '%Y_%m_%d_%H_%M_%S_%Z'
        if not timestamp:
            log_specifier = time.strftime(time_format)
        else:
            log_specifier = time.strftime(time_format, timestamp)
        log_postfix = '.h5'
        return (log_prefix + '_' + log_specifier + '_' + log_postfix)

    def record(self, data, direction=Logger.WRITE):
        '''
        Send the specified data to log file

        .. note:: buffer is flushed after all ``data`` is placed in buffer, this
            means that the buffer size will exceed the set value temporarily

        :param data: list of data to be written to log
        :param direction: ``Logger.WRITE`` if packets were sent to
            ASICs, ``Logger.READ`` if packets
            were received from ASICs. (default: ``Logger.WRITE``)

        '''
        if not self.is_enabled():
            return
        if not isinstance(data, list):
            raise ValueError('data must be a list')

        for data_obj in data:
            data_obj.direction = direction
            dataset = self.data_desc_map[type(data_obj)]
            self._buffer[dataset].append(data_obj)

        if any([len(buff) > self.buffer_length for dataset, buff in self._buffer.items()]):
            self.flush()

    def enable(self):
        '''
        Enable the logger and set up output file.

        If the file already exists then data will be appended to the end of arrays.

        :param enable: ``True`` if you want to enable the logger after
            initializing (Optional, default=``True``)
        '''
        super(HDF5Logger, self).enable()
        to_file(self.filename, [], version=self.version)

    def flush(self):
        '''
        Flushes any held data to the output file
        '''
        to_file(self.filename, self._buffer['packets'],
                version=self.version)
        self._buffer['packets'] = []




from larpix import Key

class TriggerPacket(object):
    '''
    A packet-like object which contains trigger data (e.g. associated with
    trigger words in the PACMAN ZMQ message format).

    It implements the core methods used by Packet, so it works seamlessly in
    lists of packets and in a PacketCollection.

    :param trigger_type: optional, an 8-bit type indication for the trigger type
    
    :param timestamp: optional, a 32-bit timestamp

    :param io_group: optional, an 8-bit io_group id

    '''
    packet_type = 7
    
    def __init__(self, trigger_type=None, timestamp=None, io_group=None):
        self.trigger_type = trigger_type
        self.timestamp = timestamp
        self.io_group = io_group

    def __eq__(self, other):
        if self.trigger_type != other.trigger_type: return False
        if self.timestamp != other.timestamp: return False
        if self.io_group != other.io_group: return False
        return True

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        strings = ['Trigger']
        if self.io_group is not None:
            strings.append('IO group: {}'.format(self.io_group))        
        if self.trigger_type is not None:
            strings.append('Type: {}'.format(self.trigger_type))
        if self.timestamp is not None:
            strings.append('Timestamp: {}'.format(self.timestamp))
        return '[ '+' | '.join(strings)+' ]'

    def __repr__(self):
        strings = list()
        if self.io_group is not None:
            strings.append('io_group={}'.format(self.io_group))        
        if self.trigger_type is not None:
            strings.append('sync_type={}'.format(self.trigger_type))
        if self.timestamp is not None:
            strings.append('timestamp={}'.format(self.timestamp))
        return 'TriggerPacket('+', '.join(strings)+')'

    def export(self):
        d = dict()
        d['io_group'] = self.io_group
        d['trigger_type'] = self.trigger_type
        d['timestamp'] = self.timestamp
        d['type'] = self.packet_type
        return d

    def from_dict(self, d):
        if 'io_group' in d:
            self.io_group = d['io_group']
        if 'trigger_type' in d:
            self.trigger_type = d['trigger_type']
        if 'timestamp' in d:
            self.timestamp = d['timestamp']

    @property
    def chip_key(self):
        if self.io_group:
            return Key(self.io_group,0,0)
        return None

    @chip_key.setter
    def	chip_key(self, val):
        key = Key(val)
        self.io_group = key.io_group



import warnings
warnings.simplefilter('default', DeprecationWarning)

class Logger(object):
    '''
    Base class for larpix logger objects that explicity describes the necessary
    functions for a Logger implementation. Additional functions are not built
    into the larpix core.

    '''

    #: Flag to indicate packets were sent to ASICs
    WRITE = 0
    #: Flag to indicate packets were received from ASICs
    READ = 1

    def __init__(self, enabled=False, *args, **kwargs):
        '''
        Create new logger instance.

        '''
        self._enabled = enabled
        self._open = False

    def record(self, data, direction=0, *args, **kwargs):
        '''
        Log specified data.

        :param data: ``list`` of data to be written to log. Valid data types are specified by logger implementation. Raises a ``ValueError`` if datatype is invalid.
        :param direction: ``Logger.WRITE`` if packets were sent to
            ASICs, ``Logger.READ`` if packets
            were received from ASICs. (default: ``Logger.WRITE``)

        '''
        pass

    def is_enabled(self):
        '''
        Check if logger is enabled, i.e. actively recording data.
        All data passed into ``record()`` between an ``enable()`` and
        ``disable()`` command should be reflected in the log.

        '''
        return self._enabled

    def enable(self):
        '''
        Enable logger

        '''
        self._enabled = True

    def disable(self):
        '''
        Disable logger

        .. note:: This flushes any data in the buffer before disabling

        '''
        if self._enabled:
            self.flush()
        self._enabled = False

    def flush(self):
        '''
        Flushes any held data from memory to the destination

        '''
        pass



from bitarray import bitarray
import struct

from .. import bitarrayhelper as bah
from ..key import Key

class Packet_v1(object):
    '''
    A single 54-bit LArPix UART data packet.

    LArPix Packet objects have attributes for inspecting and modifying
    the contents of the packet.

    Internally, packets are represented as an array of bits, and the
    different attributes use Python "properties" to seamlessly convert
    between the bits representation and a more intuitive integer
    representation. The bits representation can be inspected with the
    ``bits`` attribute.

    Packet objects do not restrict you from adjusting an attribute for an
    inappropriate packet type. For example, you can create a data packet and
    then set ``packet.register_address = 5``. This will adjust the packet
    bits corresponding to a configuration packet's "register\_address"
    region, which is probably not what you want for your data packet.

    Packets have a parity bit which enforces odd parity, i.e. the sum of
    all the individual bits in a packet must be an odd number. The parity
    bit can be accessed as above using the ``parity_bit_value`` attribute.
    The correct parity bit can be computed using ``compute_parity()``,
    and the validity of a packet's parity can be checked using
    ``has_valid_parity()``. When constructing a new packet, the correct
    parity bit can be assigned using ``assign_parity()``.

    Individual packets can be printed to show a human-readable
    interpretation of the packet contents. The printed version adjusts its
    output based on the packet type, so a data packet will show the data
    word, timestamp, etc., while a configuration packet will show the register
    address and register data.

    '''
    asic_version = 1
    size = 54
    num_bytes = 7
    # These ranges are reversed from the bit addresses given in the
    # LArPix datasheet because BitArray indexing is big-endian but we
    # transmit data little-endian-ly. E.g.:
    # >>> x = BitArray('0b00')
    # >>> x[0:] = bin(2)  # ('0b10')
    # >>> x[0]  # returns True (1)
    # Another way to think of it is BitArray indexing reads the
    # bitstream from 0:N but all the LArPix datasheet indexing goes from
    # N:0.
    packet_type_bits = slice(52, 54)
    chipid_bits = slice(44, 52)
    parity_bit = 0
    parity_calc_bits = slice(1, 54)
    channel_id_bits = slice(37, 44)
    timestamp_bits = slice(13, 37)
    dataword_bits = slice(3, 13)
    fifo_half_bit = 2
    fifo_full_bit = 1
    register_address_bits = slice(36, 44)
    register_data_bits = slice(28, 36)
    config_unused_bits = slice(1, 28)
    test_counter_bits_11_0 = slice(1, 13)
    test_counter_bits_15_12 = slice(40, 44)

    DATA_PACKET = bitarray('00')
    TEST_PACKET = bitarray('01')
    CONFIG_WRITE_PACKET = bitarray('10')
    CONFIG_READ_PACKET = bitarray('11')
    _bit_padding = bitarray('00')

    def __init__(self, bytestream=None):
        if bytestream is None:
            self.bits = bitarray(self.size)
            self.bits.setall(False)
            return
        elif len(bytestream) == self.num_bytes:
            # Parse the bytestream. Remember that bytestream[0] goes at
            # the 'end' of the BitArray
            reversed_bytestream = bytestream[::-1]
            self.bits = bitarray()
            self.bits.frombytes(reversed_bytestream)
            # Get rid of the padding (now at the beginning of the
            # bitstream because of the reverse order)
            self.bits = self.bits[2:]
        else:
            raise ValueError('Invalid number of bytes: %s' %
                    len(bytestream))

    def __eq__(self, other):
        return self.bits == other.bits

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        string = '[ '
        string += 'Chip key: {} | '.format(self.chip_key)
        if hasattr(self, 'direction'):
            string += 'Direction: {}'.format(self.direction)
            string += ' | '
        ptype = self.packet_type
        if ptype == Packet_v1.TEST_PACKET:
            string += 'Test | '
            string += 'Counter: %d | ' % self.test_counter
        elif ptype == Packet_v1.DATA_PACKET:
            string += 'Data | '
            string += 'Channel: %d | ' % self.channel_id
            string += 'Timestamp: %d | ' % self.timestamp
            string += 'ADC data: %d | ' % self.dataword
            string += 'FIFO Half: %s | ' % bool(self.fifo_half_flag)
            string += 'FIFO Full: %s | ' % bool(self.fifo_full_flag)
        elif (ptype == Packet_v1.CONFIG_READ_PACKET or ptype ==
                Packet_v1.CONFIG_WRITE_PACKET):
            if ptype == Packet_v1.CONFIG_READ_PACKET:
                string += 'Config read | '
            else:
                string += 'Config write | '
            string += 'Register: %d | ' % self.register_address
            string += 'Value: % d | ' % self.register_data
        first_splitter = string.find('|')
        string = (string[:first_splitter] + '| Chip: %d ' % self.chipid +
                string[first_splitter:])
        string += ('Parity: %d (valid: %s) ]' %
                (self.parity_bit_value, self.has_valid_parity()))
        return string

    def __repr__(self):
        return 'Packet_v1(' + str(self.bytes()) + ')'

    def bytes(self):
        '''
        Construct the bytes that make up the packet.

        Byte 0 is the first byte that would be sent out and contains the
        first 8 bits of the packet (i.e. packet type and part of the
        chip ID).

        *Note*: The internal bits representation of the packet has a
        different endian-ness compared to the output of this method.

        '''
        # Here's the only other place we have to deal with the
        # endianness issue by reversing the order
        padded_output = self._bit_padding + self.bits
        bytes_output = padded_output.tobytes()
        return bytes_output[::-1]

    def export(self):
        '''Return a dict representation of this Packet.'''
        type_map = {
                self.TEST_PACKET.to01(): 'test',
                self.DATA_PACKET.to01(): 'data',
                self.CONFIG_WRITE_PACKET.to01(): 'config write',
                self.CONFIG_READ_PACKET.to01(): 'config read'
                }
        d = {}
        d['asic_version'] = self.asic_version
        d['chip_key'] = str(self.chip_key) if self.chip_key else None
        d['bits'] = self.bits.to01()
        d['type_str'] = type_map[self.packet_type.to01()]
        d['type'] = bah.touint(self.packet_type)
        d['chipid'] = self.chipid
        d['parity'] = self.parity_bit_value
        d['valid_parity'] = self.has_valid_parity()
        ptype = self.packet_type
        if ptype == Packet_v1.TEST_PACKET:
            d['counter'] = self.test_counter
        elif ptype == Packet_v1.DATA_PACKET:
            d['channel'] = self.channel_id
            d['timestamp'] = self.timestamp
            d['adc_counts'] = self.dataword
            d['fifo_half'] = bool(self.fifo_half_flag)
            d['fifo_full'] = bool(self.fifo_full_flag)
        elif (ptype == Packet_v1.CONFIG_READ_PACKET or ptype ==
                Packet_v1.CONFIG_WRITE_PACKET):
            d['register'] = self.register_address
            d['value'] = self.register_data
        return d

    def from_dict(self, d):
        ''' Inverse of export - modify packet based on dict '''
        if not d['asic_version'] == self.asic_version:
            raise ValueError('invalid asic version {}'.format(d['asic_version']))
        if 'type' in d and d['type'] not in [bah.touint(packet_type) for packet_type in \
            (Packet_v1.DATA_PACKET, Packet_v1.TEST_PACKET, Packet_v1.CONFIG_WRITE_PACKET, Packet_v1.CONFIG_READ_PACKET)]:
            raise ValueError('invalid packet type for Packet_v2')
        for key, value in d.items():
            if key in ('type_str', 'valid_parity'):
                continue
            elif key == 'bits':
                self.bits = bitarray(value)
            elif key == 'type':
                self.packet_type = value
            elif key == 'register':
                self.register_address = value
            elif key == 'value':
                self.register_data = value
            elif key == 'adc_counts':
                self.dataword = value
            elif key == 'parity':
                self.parity_bit_value = value
            elif key == 'counter':
                self.test_counter = value
            elif key == 'channel':
                self.channel_id = value
            elif key == 'fifo_half':
                self.fifo_half_flag = value
            elif key == 'fifo_full':
                self.fifo_full_flag = value
            else:
                setattr(self, key, value)

    @property
    def chip_key(self):
        if hasattr(self, '_chip_key'):
            return self._chip_key
        if self.io_group is None or self.io_channel is None:
            return None
        self._chip_key = Key(self.io_group, self.io_channel, self.chipid)
        return self._chip_key

    @chip_key.setter
    def chip_key(self, value):
        # remove cached key
        if hasattr(self, '_chip_key'):
            del self._chip_key
        if value is None:
            self.io_channel = None
            self.io_group = None
            return
        if isinstance(value, Key):
            self.io_group = value.io_group
            self.io_channel = value.io_channel
            self.chipid = value.chip_id
            return
        # try again by casting as a Key
        self.chip_key = Key(value)

    @property
    def io_group(self):
        if hasattr(self, '_io_channel'):
            return self._io_group
        return None

    @io_group.setter
    def io_group(self, value):
        if hasattr(self, '_chip_key'):
            # remove cached key
            del self._chip_key
        if value is None:
            if hasattr(self, '_io_group'):
                del self._io_group
            return
        # no value validation!
        self._io_group = value

    @property
    def io_channel(self):
        if hasattr(self, '_io_channel'):
            return self._io_channel
        return None

    @io_channel.setter
    def io_channel(self, value):
        if hasattr(self, '_chip_key'):
            # remove cached key
            del self._chip_key
        if value is None:
            if hasattr(self, '_io_channel'):
                del self._io_channel
            return
        # no value validation!
        self._io_channel = value

    @property
    def packet_type(self):
        return self.bits[self.packet_type_bits]

    @packet_type.setter
    def packet_type(self, value):
        self.bits[self.packet_type_bits] = bah.fromuint(value,
                self.packet_type_bits)

    @property
    def chipid(self):
        return bah.touint(self.bits[self.chipid_bits])

    @chipid.setter
    def chipid(self, value):
        if hasattr(self, '_chip_key'):
            # remove cached key
            del self._chip_key
        self.bits[self.chipid_bits] = bah.fromuint(value,
                self.chipid_bits)

    @property
    def chip_id(self):
        # additional handle to match Packet_v2 name
        return self.chipid

    @chip_id.setter
    def chip_id(self, value):
        self.chipid = value

    @property
    def parity_bit_value(self):
        return int(self.bits[self.parity_bit])

    @parity_bit_value.setter
    def parity_bit_value(self, value):
        self.bits[self.parity_bit] = bool(value)

    def compute_parity(self):
        return 1 - (self.bits[self.parity_calc_bits].count(True) % 2)

    def assign_parity(self):
        self.parity_bit_value = self.compute_parity()

    def has_valid_parity(self):
        return self.parity_bit_value == self.compute_parity()

    @property
    def channel_id(self):
        return bah.touint(self.bits[self.channel_id_bits])

    @channel_id.setter
    def channel_id(self, value):
        self.bits[self.channel_id_bits] = bah.fromuint(value,
                self.channel_id_bits)

    @property
    def timestamp(self):
        return bah.touint(self.bits[self.timestamp_bits])

    @timestamp.setter
    def timestamp(self, value):
        self.bits[self.timestamp_bits] = bah.fromuint(value,
                self.timestamp_bits)

    @property
    def dataword(self):
        ostensible_value = bah.touint(self.bits[self.dataword_bits])
        # TODO fix in LArPix v2
        return ostensible_value - (ostensible_value % 2)

    @dataword.setter
    def dataword(self, value):
        self.bits[self.dataword_bits] = bah.fromuint(value,
                self.dataword_bits)

    @property
    def fifo_half_flag(self):
        return int(self.bits[self.fifo_half_bit])

    @fifo_half_flag.setter
    def fifo_half_flag(self, value):
        self.bits[self.fifo_half_bit] = bool(value)

    @property
    def fifo_full_flag(self):
        return int(self.bits[self.fifo_full_bit])

    @fifo_full_flag.setter
    def fifo_full_flag(self, value):
        self.bits[self.fifo_full_bit] = bool(value)

    @property
    def register_address(self):
        return bah.touint(self.bits[self.register_address_bits])

    @register_address.setter
    def register_address(self, value):
        self.bits[self.register_address_bits] = bah.fromuint(value,
                self.register_address_bits)

    @property
    def register_data(self):
        return bah.touint(self.bits[self.register_data_bits])

    @register_data.setter
    def register_data(self, value):
        self.bits[self.register_data_bits] = bah.fromuint(value,
                self.register_data_bits)

    @property
    def test_counter(self):
        return bah.touint(self.bits[self.test_counter_bits_15_12] +
                self.bits[self.test_counter_bits_11_0])

    @test_counter.setter
    def test_counter(self, value):
        allbits = bah.fromuint(value, 16)
        self.bits[self.test_counter_bits_15_12] = (
            bah.fromuint(allbits[:4], self.test_counter_bits_15_12))
        self.bits[self.test_counter_bits_11_0] = (
            bah.fromuint(allbits[4:], self.test_counter_bits_11_0))



'''
Tests for larpix.io.multizmq_io module

'''
from __future__ import print_function
import pytest
import os
import json
from larpix.larpix import Packet, Key, Packet_v1
from larpix.io.multizmq_io import MultiZMQ_IO
from larpix.format.message_format import dataserver_message_decode, dataserver_message_encode

@pytest.fixture
def io_config(tmpdir):
    filename = str(tmpdir.join('test_conf.json'))
    config_dict = {
            "_config_type": "io",
            "io_class": "MultiZMQ_IO",
            "io_group": [
                [1, "192.0.2.1"],
                [2, "192.0.2.2"]
            ]
        }
    with open(filename,'w') as of:
        json.dump(config_dict, of)
    return filename

@pytest.fixture
def multizmq_io_obj(io_config):
    return MultiZMQ_IO(io_config)

def test_encode(multizmq_io_obj):
    chip_id = 64
    io_chain = 1
    io_group = list(multizmq_io_obj._io_group_table)[0]
    test_packet = Packet(b'\x01'*Packet.num_bytes)
    test_packet.io_group = io_group
    test_packet.io_channel = io_chain
    test_bytes = b'0x0101010101010101 1'
    if Packet == Packet_v1:
        test_bytes = b'0x0001010101010101 1'
    expected = [test_bytes]
    assert multizmq_io_obj.encode([test_packet]) == expected

def test_decode(multizmq_io_obj):
    chip_id = 64
    io_chain = 1
    io_group = list(multizmq_io_obj._io_group_table)[0]
    address = str(multizmq_io_obj._io_group_table[io_group])
    test_packet = Packet(b'\x01'*Packet.num_bytes)
    test_packet.io_group = io_group
    test_packet.io_channel = io_chain
    test_bytes = dataserver_message_encode([test_packet])
    expected = [test_packet]
    assert multizmq_io_obj.decode(test_bytes, address=address) == expected



'''
This module gives access to the LArPix+HDF5 file format.

File format description
=======================

All LArPix+HDF5 files use the HDF5 format so that they
can be read and written using any language that has an HDF5 binding. The
documentation for the Python h5py binding is at <http://docs.h5py.org>.

The ``to_file`` and ``from_file`` methods translate between a list of
Packet-like objects and an HDF5 data file. ``from_file`` can be used to
load up the full file all at once or just a subset of rows (supposing
the full file was too big to fit in memory). To access the data most
efficiently, do not rely on ``from_file`` and instead perform analysis
directly on the HDF5 data file.

File Header
-----------

The file header can be found in the ``/_header`` HDF5 group. At a
minimum, the header will contain the following HDF5 attributes:

    - ``version``: a string containing the LArPix+HDF5 version
    - ``created``: a Unix timestamp of the file's creation time
    - ``modified``: a Unix timestamp of the file's last-modified time

Versions
--------

The LArPix+HDF5 format is self-describing and versioned. This means as
the format evolves, the files themselves will identify which version
of the format should be used to interpret them. When writing a file
with ``to_file``, the format version can be specified, or by default,
the latest version is used. When reading a file with ``from_file``, by
default, the format version of the actual file is used. If a specific
format version is expected or required, that version can be specified,
and a ``RuntimeError`` will be raised if a different format version is
encountered.

The versions are always in the format ``major.minor`` and are stored as
strings (e.g. ``'1.0'``, ``'1.5'``, ``2.0``).

The minor format will increase if a non-breaking change is made, so that
a script compatible with a lower minor version will also work with files
that have a higher minor version. E.g. a script designed to work with
v1.0 will also work with v1.5. The reverse is not necessarily true: a
script designed to work with v1.5 *may not* work with v1.0 files.

The major format will increase if a breaking change is made. This means
that a script designed to work with v1.5 will *likely not* work with
v2.0 files, and vice versa.

File Data
---------

The file data is saved in HDF5 datasets, and the specific data format
depends on the LArPix+HDF5 version.

Version 2.2 description
^^^^^^^^^^^^^^^^^^^^^^^

For version 2.2, there are two dataset: ``packets`` and ``messages``.

The ``packets`` dataset
contains a list of all of the packets sent and received during a
particular time interval.

    - Shape: ``(N,)``, ``N >= 0``

    - Datatype: a compound datatype (called "structured type" in
      h5py/numpy). Not all fields are relevant for each packet. Unused
      fields are set to a default value of 0 or the empty string.
      Keys/fields:

        - ``io_group`` (``u1``/unsigned byte): an id associated with the
          high-level io group associated with this packet

        - ``io_channel`` (``u1``/unsigned byte): the id associated with the
          mid-level io channel associated with this packet

        - ``packet_type`` (``u1``/unsigned byte): the packet type code, which
          can be interpreted according to the map stored in the
          'packets' attribute 'packet_types'

        - ``chip_id`` (``u1``/unsigned byte): the LArPix chip id

        - ``parity`` (``u1``/unsigned byte): the packet parity bit (0 or
          1)

        - ``valid_parity`` (``u1``/unsigned byte): 1 if the packet
          parity is valid (odd), 0 if it is invalid

        - ``downstream_marker`` (``u1``/unsigned byte): a marker to indicate the
          hydra io network direction for this packet

        - ``channel_id`` (``u1``/unsigned byte): the ASIC channel

        - ``timestamp`` (``u8``/unsigned 8-byte long int): the timestamp
          associated with the packet. **Caution**: this field does
          "many-duty" as both the ASIC timestamp in data packets
          (``type`` == 0), as the global timestamp in timestamp
          packets (``type`` == 4), as the message timestamp in
          message packets (``type`` == 5), as the timestamp in sync packets
          (``type`` == 6), and the timestamp in trigger packets (``type`` == 7).

        - ``first_packet`` (``u1``/unsigned byte): indicates if this is the
          packet recieved in a trigger burst (v2.1 or newer only)

        - ``dataword`` (``u1``/unsigned byte): the ADC data word
          **Caution**: as of v2.2, this field does double duty as both the 
          LArPix ADC value in data packets (``type`` == 0) and the clk source
          value in sync packets (``type`` == 6).

        - ``trigger_type`` (``u1``/unsigned byte): the trigger type assciated
          with this packet **Caution**: as of v2.2, this field does triple
          duty as both the LArPix packet trigger type in data packets 
          (``type`` == 0), the sync type in sync packets (``type`` == 6), and
          the trigger type in trigger packets (``type`` == 7).

        - ``local_fifo` (``u1``/unsigned byte): 1 if the channel FIFO is >50%
          full, 3 if the channel FIFO is 100% full

        - ``shared_fifo`` (``u1``/unsigned byte): 1 if the chip FIFO is >50%
          full, 3 if the channel FIFO is 100% full

        - ``register_address`` (``u1``/unsigned byte): the configuration
          register index

        - ``register_data`` (``u1``/unsigned byte): the configuration register
          value

        - ``direction`` (``u1``/unsigned byte): 0 if packet was sent to
          ASICs, 1 if packet was received from ASICs.

        - ``local_fifo_events`` (``u1``/unsigned byte): number of packets in the
          channel FIFO (only valid if FIFO diagnostics are enabled)

        - ``shared_fifo_events`` (``u2``/unsigned byte): number of packets in the
          chip FIFO (only valid if FIFO diagnostics are enabled)

        - ``counter`` (``u4``/unsigned 4-byte int): the message index (only
          valid for message type packets)

        - ``fifo_diagnostics_enabled`` (``u1``/unsigned byte): flag for when
          fifo diagnostics are enabled (1 if enabled, 0 if not)


    - Packet types lookup: the ``packets`` dataset has an attribute
      ``'packet_types'`` which contains the following lookup table for
      packets::

        0: 'data',
        1: 'test',
        2: 'config write',
        3: 'config read',
        4: 'timestamp',
        5: 'message',
        6: 'sync',
        7: 'trigger'

The ``messages`` dataset has the full messages referred to by message
packets in the ``packets`` dataset.

    - Shape: ``(N,)``, ``N >= 0``

    - Datatype: a compound datatype with fields:

        - ``message`` (``S64``/64-character string): the message

        - ``timestamp`` (``u8``/unsigned 8-byte long int): the timestamp
          associated with the message

        - ``index`` (``u4``/unsigned 4-byte int): the message index,
          which should be equal to the row index in the ``messages``
          dataset


Version 2.1 description
^^^^^^^^^^^^^^^^^^^^^^^

For version 2.1, there are two dataset: ``packets`` and ``messages``.

The ``packets`` dataset
contains a list of all of the packets sent and received during a
particular time interval.

    - Shape: ``(N,)``, ``N >= 0``

    - Datatype: a compound datatype (called "structured type" in
      h5py/numpy). Not all fields are relevant for each packet. Unused
      fields are set to a default value of 0 or the empty string.
      Keys/fields:

        - ``io_group`` (``u1``/unsigned byte): an id associated with the
          high-level io group associated with this packet

        - ``io_channel`` (``u1``/unsigned byte): the id associated with the
          mid-level io channel associated with this packet

        - ``packet_type`` (``u1``/unsigned byte): the packet type code, which
          can be interpreted according to the map stored in the
          'packets' attribute 'packet_types'

        - ``chip_id`` (``u1``/unsigned byte): the LArPix chip id

        - ``parity`` (``u1``/unsigned byte): the packet parity bit (0 or
          1)

        - ``valid_parity`` (``u1``/unsigned byte): 1 if the packet
          parity is valid (odd), 0 if it is invalid

        - ``downstream_marker`` (``u1``/unsigned byte): a marker to indicate the
          hydra io network direction for this packet

        - ``channel_id`` (``u1``/unsigned byte): the ASIC channel

        - ``timestamp`` (``u8``/unsigned 8-byte long int): the timestamp
          associated with the packet. **Caution**: this field does
          "triple duty" as both the ASIC timestamp in data packets
          (``type`` == 0), as the global timestamp in timestamp
          packets (``type`` == 4), and as the message timestamp in
          message packets (``type`` == 5).

        - ``first_packet`` (``u1``/unsigned byte): indicates if this is the
          packet recieved in a trigger burst (v2.1 or newer only)

        - ``dataword`` (``u1``/unsigned byte): the ADC data word

        - ``trigger_type`` (``u1``/unsigned byte): the trigger type assciated
          with this packet

        - ``local_fifo` (``u1``/unsigned byte): 1 if the channel FIFO is >50%
          full, 3 if the channel FIFO is 100% full

        - ``shared_fifo`` (``u1``/unsigned byte): 1 if the chip FIFO is >50%
          full, 3 if the channel FIFO is 100% full

        - ``register_address`` (``u1``/unsigned byte): the configuration
          register index

        - ``register_data`` (``u1``/unsigned byte): the configuration register
          value

        - ``direction`` (``u1``/unsigned byte): 0 if packet was sent to
          ASICs, 1 if packet was received from ASICs.

        - ``local_fifo_events`` (``u1``/unsigned byte): number of packets in the
          channel FIFO (only valid if FIFO diagnostics are enabled)

        - ``shared_fifo_events`` (``u2``/unsigned byte): number of packets in the
          chip FIFO (only valid if FIFO diagnostics are enabled)

        - ``counter`` (``u4``/unsigned 4-byte int): the message index (only
          valid for message type packets)

        - ``fifo_diagnostics_enabled`` (``u1``/unsigned byte): flag for when
          fifo diagnostics are enabled (1 if enabled, 0 if not)

    - Packet types lookup: the ``packets`` dataset has an attribute
      ``'packet_types'`` which contains the following lookup table for
      packets::

        0: 'data',
        1: 'test',
        2: 'config write',
        3: 'config read',
        4: 'timestamp',
        5: 'message',

The ``messages`` dataset has the full messages referred to by message
packets in the ``packets`` dataset.

    - Shape: ``(N,)``, ``N >= 0``

    - Datatype: a compound datatype with fields:

        - ``message`` (``S64``/64-character string): the message

        - ``timestamp`` (``u8``/unsigned 8-byte long int): the timestamp
          associated with the message

        - ``index`` (``u4``/unsigned 4-byte int): the message index,
          which should be equal to the row index in the ``messages``
          dataset


Version 1.0 description
^^^^^^^^^^^^^^^^^^^^^^^

For version 1.0, there are two dataset: ``packets`` and ``messages``.

The ``packets`` dataset
contains a list of all of the packets sent and received during a
particular time interval.

    - Shape: ``(N,)``, ``N >= 0``

    - Datatype: a compound datatype (called "structured type" in
      h5py/numpy). Not all fields are relevant for each packet. Unused
      fields are set to a default value of 0 or the empty string.
      Keys/fields:

        - ``chip_key`` (``S32``/32-character string): the chip key
          identifying the ASIC associated with this packet

        - ``type`` (``u1``/unsigned byte): the packet type code, which
          can be interpreted according to the map stored in the
          raw_packet attribute 'packet_types'

        - ``chipid`` (``u1``/unsigned byte): the LArPix chipid

        - ``parity`` (``u1``/unsigned byte): the packet parity bit (0 or
          1)

        - ``valid_parity`` (``u1``/unsigned byte): 1 if the packet
          parity is valid (odd), 0 if it is invalid

        - ``channel`` (``u1``/unsigned byte): the ASIC channel

        - ``timestamp`` (``u8``/unsigned 8-byte long int): the timestamp
          associated with the packet. **Caution**: this field does
          "triple duty" as both the ASIC timestamp in data packets
          (``type`` == 0), as the global timestamp in timestamp
          packets (``type`` == 4), and as the message timestamp in
          message packets (``type`` == 5).

        - ``adc_counts`` (``u1``/unsigned byte): the ADC data word

        - ``fifo_half`` (``u1``/unsigned byte): 1 if the FIFO half full
          flag is present, 0 otherwise.

        - ``fifo_full`` (``u1``/unsigned byte): 1 if the FIFO full flag
          is present, 0 otherwise.

        - ``register`` (``u1``/unsigned byte): the configuration
          register index

        - ``value`` (``u1``/unsigned byte): the configuration register
          value

        - ``counter`` (``u4``/unsigned 4-byte int): the test counter
          value, or the message index. **Caution**: this field does
          "double duty" as the counter for test packets (``type`` == 1)
          and as the message index for message packets (``type`` == 5).

        - ``direction`` (``u1``/unsigned byte): 0 if packet was sent to
          ASICs, 1 if packet was received from ASICs.

    - Packet types lookup: the ``packets`` dataset has an attribute
      ``'packet_types'`` which contains the following lookup table for
      packets::

        0: 'data',
        1: 'test',
        2: 'config write',
        3: 'config read',
        4: 'timestamp',
        5: 'message',

The ``messages`` dataset has the full messages referred to by message
packets in the ``packets`` dataset.

    - Shape: ``(N,)``, ``N >= 0``

    - Datatype: a compound datatype with fields:

        - ``message`` (``S64``/64-character string): the message

        - ``timestamp`` (``u8``/unsigned 8-byte long int): the timestamp
          associated with the message

        - ``index`` (``u4``/unsigned 4-byte int): the message index,
          which should be equal to the row index in the ``messages``
          dataset

Examples
--------

Plot a histogram of ADC counts (selecting packet type to be data packets
only)

>>> import matplotlib.pyplot as plt
>>> import h5py
>>> f = h5py.File('output.h5', 'r')
>>> packets = f['packets']
>>> plt.hist(packets['adc_counts'][packets['type'] == 0])
>>> plt.show()

Load the first 10 packets in a file into Packet objects and print any
MessagePacket packets to the console

>>> from larpix.format.hdf5format import from_file
>>> from larpix.larpix import MessagePacket
>>> result = from_file('output.h5', end=10)
>>> for packet in result['packets']:
...     if isinstance(packet, MessagePacket):
...         print(packet)



'''
import time

import h5py
import numpy as np
import struct

from larpix.larpix import Packet_v1, Packet_v2, TimestampPacket, MessagePacket, SyncPacket, TriggerPacket
from larpix.logger import Logger

#: The most recent / up-to-date LArPix+HDF5 format version
latest_version = '2.3'

#: The dtype specification used in the HDF5 files.
#:
#: Structure: ``{version: {dset_name: [structured dtype fields]}}``
dtypes = {
        '0.0': {
            'raw_packet': [
                ('chip_key','S32'),
                ('type','u1'),
                ('chipid','u1'),
                ('parity','u1'),
                ('valid_parity','u1'),
                ('counter','u4'),
                ('channel','u1'),
                ('timestamp','u8'),
                ('adc_counts','u1'),
                ('fifo_half','u1'),
                ('fifo_full','u1'),
                ('register','u1'),
                ('value','u1'),
                ]
            },
        '1.0': { # compatible with v1 packets only
            'packets': [
                ('chip_key','S32'),
                ('type','u1'),
                ('chipid','u1'),
                ('parity','u1'),
                ('valid_parity','u1'),
                ('channel','u1'),
                ('timestamp','u8'),
                ('adc_counts','u1'),
                ('fifo_half','u1'),
                ('fifo_full','u1'),
                ('register','u1'),
                ('value','u1'),
                ('counter','u4'),
                ('direction', 'u1'),
                ],
            'messages': [
                ('message', 'S64'),
                ('timestamp', 'u8'),
                ('index', 'u4'),
                ]
            },
        '2.0': { # compatible with v2 packets and timestamp packets only
            'packets': [
                ('io_group','u1'),
                ('io_channel','u1'),
                ('chip_id','u1'),
                ('packet_type','u1'),
                ('downstream_marker','u1'),
                ('parity','u1'),
                ('valid_parity','u1'),
                ('channel_id','u1'),
                ('timestamp','u8'),
                ('dataword','u1'),
                ('trigger_type','u1'),
                ('local_fifo','u1'),
                ('shared_fifo','u1'),
                ('register_address','u1'),
                ('register_data','u1'),
                ('direction', 'u1'),
                ('local_fifo_events','u1'),
                ('shared_fifo_events','u2'),
                ('counter','u4'),
                ('fifo_diagnostics_enabled','u1'),
                ],
            'messages': [
                ('message', 'S64'),
                ('timestamp', 'u8'),
                ('index', 'u4'),
                ]
            },
        '2.1': { # compatible with v2 packets and timestamp packets only
            'packets': [
                ('io_group','u1'),
                ('io_channel','u1'),
                ('chip_id','u1'),
                ('packet_type','u1'),
                ('downstream_marker','u1'),
                ('parity','u1'),
                ('valid_parity','u1'),
                ('channel_id','u1'),
                ('timestamp','u8'),
                ('dataword','u1'),
                ('trigger_type','u1'),
                ('local_fifo','u1'),
                ('shared_fifo','u1'),
                ('register_address','u1'),
                ('register_data','u1'),
                ('direction', 'u1'),
                ('local_fifo_events','u1'),
                ('shared_fifo_events','u1'),
                ('counter','u4'),
                ('fifo_diagnostics_enabled','u1'),
                ('first_packet','u1'),                
                ],
            'messages': [
                ('message', 'S64'),
                ('timestamp', 'u8'),
                ('index', 'u4'),
                ]
            },
        '2.2': { # compatible with v2 packets, timestamp packets,
                 # sync packets, and trigger packets only
            'packets': [
                ('io_group','u1'),
                ('io_channel','u1'),
                ('chip_id','u1'),
                ('packet_type','u1'),
                ('downstream_marker','u1'),
                ('parity','u1'),
                ('valid_parity','u1'),
                ('channel_id','u1'),
                ('timestamp','u8'),
                ('dataword','u1'),
                ('trigger_type','u1'),
                ('local_fifo','u1'),
                ('shared_fifo','u1'),
                ('register_address','u1'),
                ('register_data','u1'),
                ('direction', 'u1'),
                ('local_fifo_events','u1'),
                ('shared_fifo_events','u1'),
                ('counter','u4'),
                ('fifo_diagnostics_enabled','u1'),
                ('first_packet','u1'),                
                ],
            'messages': [
                ('message', 'S64'),
                ('timestamp', 'u8'),
                ('index', 'u4'),
                ]
            },
            '2.3': { # compatible with v2 packets, timestamp packets,
                 # sync packets, and trigger packets only
            'packets': [
                ('io_group','u1'),
                ('io_channel','u1'),
                ('chip_id','u1'),
                ('packet_type','u1'),
                ('downstream_marker','u1'),
                ('parity','u1'),
                ('valid_parity','u1'),
                ('channel_id','u1'),
                ('timestamp','u8'),
                ('dataword','u1'),
                ('trigger_type','u1'),
                ('local_fifo','u1'),
                ('shared_fifo','u1'),
                ('register_address','u1'),
                ('register_data','u1'),
                ('direction', 'u1'),
                ('local_fifo_events','u1'),
                ('shared_fifo_events','u1'),
                ('counter','u4'),
                ('fifo_diagnostics_enabled','u1'),
                ('first_packet','u1'),
                ('receipt_timestamp','u4')
                ],
            'messages': [
                ('message', 'S64'),
                ('timestamp', 'u8'),
                ('index', 'u4'),
                ]
            }
        }

#: A map between attribute name and "column index" in the structured
#: dtypes.
#:
#: Structure: ``{version: {dset_name: {field_name: index}}}``
dtype_property_index_lookup = {
        '0.0': {
            'raw_packet': {
                'chip_key': 0,
                'type': 1,
                'chipid': 2,
                'parity': 3,
                'valid_parity': 4,
                'counter': 5,
                'channel': 6,
                'timestamp': 7,
                'adc_counts': 8,
                'fifo_half': 9,
                'fifo_full': 10,
                'register': 11,
                'value': 12,
                }
            },
        '1.0': {
            'packets': {
                'chip_key': 0,
                'type': 1,
                'chipid': 2,
                'parity': 3,
                'valid_parity': 4,
                'channel': 5,
                'timestamp': 6,
                'adc_counts': 7,
                'fifo_half': 8,
                'fifo_full': 9,
                'register': 10,
                'value': 11,
                'counter': 12,
                'direction': 13,
                },
            'messages': {
                'message': 0,
                'timestamp': 1,
                'index': 2,
                }
            },
        '2.0': {
            'packets': {
                'io_group': 0,
                'io_channel': 1,
                'chip_id': 2,
                'packet_type': 3,
                'downstream_marker': 4,
                'parity': 5,
                'valid_parity': 6,
                'channel_id': 7,
                'timestamp': 8,
                'dataword': 9,
                'trigger_type': 10,
                'local_fifo': 11,
                'shared_fifo': 12,
                'register_address': 13,
                'register_data': 14,
                'direction': 15,
                'local_fifo_events': 16,
                'shared_fifo_events': 17,
                'counter': 18,
                'fifo_diagnostics_enabled': 19,
                },
            'messages': {
                'message': 0,
                'timestamp': 1,
                'index': 2,
                }
            },
        '2.1': {
            'packets': {
                'io_group': 0,
                'io_channel': 1,
                'chip_id': 2,
                'packet_type': 3,
                'downstream_marker': 4,
                'parity': 5,
                'valid_parity': 6,
                'channel_id': 7,
                'timestamp': 8,
                'dataword': 9,
                'trigger_type': 10,
                'local_fifo': 11,
                'shared_fifo': 12,
                'register_address': 13,
                'register_data': 14,
                'direction': 15,
                'local_fifo_events': 16,
                'shared_fifo_events': 17,
                'counter': 18,
                'fifo_diagnostics_enabled': 19,
                'first_packet': 20,
                },
            'messages': {
                'message': 0,
                'timestamp': 1,
                'index': 2,
                }
        },
    '2.2': {
        'packets': {
                'io_group': 0,
                'io_channel': 1,
                'chip_id': 2,
                'packet_type': 3,
                'downstream_marker': 4,
                'parity': 5,
                'valid_parity': 6,
                'channel_id': 7,
                'timestamp': 8,
                'dataword': 9,
                'trigger_type': 10,
                'local_fifo': 11,
                'shared_fifo': 12,
                'register_address': 13,
                'register_data': 14,
                'direction': 15,
                'local_fifo_events': 16,
                'shared_fifo_events': 17,
                'counter': 18,
                'fifo_diagnostics_enabled': 19,
                'first_packet': 20,
                },
            'messages': {
                'message': 0,
                'timestamp': 1,
                'index': 2,
                }
    },
    '2.3': {
            'packets': {
                'io_group': 0,
                'io_channel': 1,
                'chip_id': 2,
                'packet_type': 3,
                'downstream_marker': 4,
                'parity': 5,
                'valid_parity': 6,
                'channel_id': 7,
                'timestamp': 8,
                'dataword': 9,
                'trigger_type': 10,
                'local_fifo': 11,
                'shared_fifo': 12,
                'register_address': 13,
                'register_data': 14,
                'direction': 15,
                'local_fifo_events': 16,
                'shared_fifo_events': 17,
                'counter': 18,
                'fifo_diagnostics_enabled': 19,
                'first_packet': 20,
                'receipt_timestamp': 21,
                },
            'messages': {
                'message': 0,
                'timestamp': 1,
                'index': 2,
                }
            }
        }

def _format_raw_packet_v0_0(pkt, version='0.0', dset='raw_packet', *args, **kwargs):
    dict_rep = pkt.export()
    encoded_packet = [
        dict_rep.get(key, b'') if val_type[0] == 'S'  # string
        else dict_rep.get(key, 0) for key, val_type in
        dtypes[version][dset]]
    return encoded_packet

def _parse_raw_packet_v0_0(row, message_dset, *args, **kwargs):
    if row['type'] == 4:
        return TimestampPacket(row['timestamp'])
    if row['type'] < 4:
        p = Packet_v1()
        p.chip_key = row['chip_key']
        p.packet_type = row['type']
        p.chipid = row['chipid']
        p.parity_bit_value = row['parity']
        if p.packet_type == Packet_v1.DATA_PACKET:
            p.channel = row['channel']
            p.timestamp = row['timestamp']
            p.dataword = row['adc_counts']
            p.fifo_half_flag = row['fifo_half']
            p.fifo_full_flag = row['fifo_full']
        elif p.packet_type == Packet_v1.TEST_PACKET:
            p.counter = row['counter']
        elif (p.packet_type == Packet_v1.CONFIG_WRITE_PACKET
              or p.packet_type == Packet_v1.CONFIG_READ_PACKET):
            p.register_address = row['register']
            p.register_data = row['value']
        return p
    return None

def _format_packets_packet_v1_0(pkt, version='1.0', dset='packets', *args, **kwargs):
    encoded_packet = _format_raw_packet_v0_0(pkt, *args, version=version, dset=dset, **kwargs)
    if hasattr(pkt, 'direction'):
        encoded_packet[dtype_property_index_lookup[version][dset]['direction']] = {
            Logger.WRITE: 0,
            Logger.READ: 1}[pkt.direction]
    return encoded_packet

def _format_messages_message_packet_v1_0(pkt, counter=0, *args, **kwargs):
    return (pkt.message, pkt.timestamp, counter)

def _parse_packets_v1_0(row, message_dset, *args, **kwargs):
    if row['type'] == 4:
        return TimestampPacket(row['timestamp'])
    if row['type'] == 5:
        index = row['counter']
        message_row = message_dset[index]
        message = message_row['message'].decode()
        timestamp = row['timestamp']
        return MessagePacket(message, timestamp)
    if row['type'] < 4:
        p = Packet_v1()
        p.chip_key = row['chip_key']
        p.packet_type = row['type']
        p.chipid = row['chipid']
        p.parity_bit_value = row['parity']
        if p.packet_type == Packet_v1.DATA_PACKET:
            p.channel = row['channel']
            p.timestamp = row['timestamp']
            p.dataword = row['adc_counts']
            p.fifo_half_flag = row['fifo_half']
            p.fifo_full_flag = row['fifo_full']
        elif p.packet_type == Packet_v1.TEST_PACKET:
            p.counter = row['counter']
        elif (p.packet_type == Packet_v1.CONFIG_WRITE_PACKET
              or p.packet_type == Packet_v1.CONFIG_READ_PACKET):
            p.register_address = row['register']
            p.register_data = row['value']
        p.direction = row['direction']
        return p
    return None

def _format_packets_packet_v2_0(pkt, version='2.0', dset='packets', *args, **kwargs):
    encoded_packet = _format_packets_packet_v1_0(pkt, version=version, dset=dset)
    if encoded_packet is not None:
        encoded_packet[dtype_property_index_lookup[version][dset]['packet_type']] = pkt.packet_type
        if isinstance(pkt, Packet_v2) and pkt.fifo_diagnostics_enabled:
            encoded_packet[dtype_property_index_lookup[version][dset]['fifo_diagnostics_enabled']] = 1
    return encoded_packet

def _parse_packets_v2_0(row, message_dset, *args, **kwargs):
    if row['packet_type'] == 4:
        return TimestampPacket(row['timestamp'])
    if row['packet_type'] == 5:
        index = row['counter']
        message_row = message_dset[index]
        message = message_row['message'].decode()
        timestamp = row['timestamp']
        return MessagePacket(message, timestamp)
    if row['packet_type'] < 4:
        p = Packet_v2()
        p.io_group = row['io_group']
        p.io_channel = row['io_channel']
        p.chip_id = row['chip_id']
        p.packet_type = row['packet_type']
        p.downstream_marker = row['downstream_marker']
        p.parity = row['parity']
        p.valid_parity = row['valid_parity']
        p.direction = row['direction']
        if p.packet_type == Packet_v2.DATA_PACKET:
            p.channel_id = row['channel_id']
            p.timestamp = row['timestamp']
            p.dataword = row['dataword']
            p.trigger_type = row['trigger_type']
            p.local_fifo = row['local_fifo']
            p.shared_fifo = row['shared_fifo']
            if row['fifo_diagnostics_enabled'] != 0:
                p.fifo_diagnostics_enabled = True
                p.local_fifo = row['local_fifo_events']
                p.shared_fifo = row['shared_fifo_events']
                p.timestamp = row['timestamp']
        elif p.packet_type in (Packet_v2.CONFIG_READ_PACKET, Packet_v2.CONFIG_WRITE_PACKET):
            p.register_address = row['register_address']
            p.register_data = row['register_data']
        return p
    return None

def _format_packets_packet_v2_1(pkt, version='2.1', dset='packets', *args, **kwargs):
    return _format_packets_packet_v2_0(pkt, *args, version=version, dset=dset, **kwargs)

def _parse_packets_v2_1(row, message_dset, *args, **kwargs):
    p = _parse_packets_v2_0(row, message_dset, *args, **kwargs)
    if isinstance(p, Packet_v2):
        p.first_packet = row['first_packet']
    return p

_uint8_struct = struct.Struct("<B")
def _format_packets_packet_v2_2(pkt, version='2.2', dset='packets', *args, **kwargs):
    encoded_packet = _format_packets_packet_v2_0(pkt, *args, version=version, dset=dset, **kwargs)
    if isinstance(pkt, SyncPacket):
        encoded_packet[dtype_property_index_lookup[version]['packets']['trigger_type']] = _uint8_struct.unpack(pkt.sync_type)[0]
        encoded_packet[dtype_property_index_lookup[version]['packets']['dataword']] = pkt.clk_source
    elif isinstance(pkt, TriggerPacket):
        encoded_packet[dtype_property_index_lookup[version]['packets']['trigger_type']] = _uint8_struct.unpack(pkt.trigger_type)[0]
    return encoded_packet

def _parse_packets_v2_2(row, message_dset, *args, **kwargs):
    p = _parse_packets_v2_1(row, message_dset, *args, **kwargs)
    if p is None:
        if row['packet_type'] == 6:
            return SyncPacket(
                io_group = row['io_group'],
                sync_type = _uint8_struct.pack(row['trigger_type']),
                clk_source = row['dataword'],
                timestamp = row['timestamp']
            )                
        if row['packet_type'] == 7:
            return TriggerPacket(
                io_group = row['io_group'],
                trigger_type = _uint8_struct.pack(row['trigger_type']),
                timestamp = row['timestamp']
            )
    return p

def _format_packets_packet_v2_3(pkt, version='2.3', dset='packets', *args, **kwargs):
    return _format_packets_packet_v2_2(pkt, *args, version=version, dset=dset, **kwargs)

def _parse_packets_v2_3(row, message_dset, *args, **kwargs):
    p = _parse_packets_v2_2(row, message_dset, *args, **kwargs)
    if isinstance(p, Packet_v2):
        p.receipt_timestamp = row['receipt_timestamp']
    return p
    

# A map between packet class and the formatting method used to convert to structured
# dtypes.
#
# Structure: ``{version: {dset_name: {larpix_packet_class: format_method}}}``
_format_method_lookup = {
    '0.0': {
        'raw_packet': {
            Packet_v1: _format_raw_packet_v0_0,
            TimestampPacket: _format_raw_packet_v0_0
        }
    },
    '1.0': {
        'packets': {
            Packet_v1: _format_packets_packet_v1_0,
            TimestampPacket: _format_packets_packet_v1_0,
            MessagePacket: _format_packets_packet_v1_0
        },
        'messages': {
            MessagePacket: _format_messages_message_packet_v1_0
        }
    },
    '2.0': {
        'packets': {
            Packet_v2: _format_packets_packet_v2_0,
            TimestampPacket: _format_packets_packet_v2_0,
            MessagePacket: _format_packets_packet_v2_0
        },
        'messages': {
            MessagePacket: _format_messages_message_packet_v1_0
        }
    },
    '2.1': {
        'packets': {
            Packet_v2: _format_packets_packet_v2_1,
            TimestampPacket: _format_packets_packet_v2_1,
            MessagePacket: _format_packets_packet_v2_1
        },
        'messages': {
            MessagePacket: _format_messages_message_packet_v1_0
        }
    },
    '2.2': {
        'packets': {
            Packet_v2: _format_packets_packet_v2_1,
            TimestampPacket: _format_packets_packet_v2_1,
            MessagePacket: _format_packets_packet_v2_1,
            SyncPacket: _format_packets_packet_v2_2,
            TriggerPacket: _format_packets_packet_v2_2
        },
        'messages': {
            MessagePacket: _format_messages_message_packet_v1_0
        },
    },
    '2.3': {
        'packets': {
            Packet_v2: _format_packets_packet_v2_3,
            TimestampPacket: _format_packets_packet_v2_3,
            MessagePacket: _format_packets_packet_v2_3,
            SyncPacket: _format_packets_packet_v2_3,
            TriggerPacket: _format_packets_packet_v2_3
        },
        'messages': {
            MessagePacket: _format_messages_message_packet_v1_0
        }       
    }
}

# A map between dset the parsing method used to convert from structured
# dtypes.
#
# Structure: ``{version: {dset_name: {larpix_packet_class: parse_method}}}``
_parse_method_lookup = {
    '0.0': {
        'raw_packet': _parse_raw_packet_v0_0
    },
    '1.0': {
        'packets': _parse_packets_v1_0
    },
    '2.0': {
        'packets': _parse_packets_v2_0
    },
    '2.1': {
        'packets': _parse_packets_v2_1
    },
    '2.2': {
        'packets': _parse_packets_v2_2
    },
    '2.3': {
        'packets': _parse_packets_v2_3
    }    
}

def to_file(filename, packet_list, mode='a', version=None):
    '''
    Save the given packets to the given file.

    This method can be used to update an existing file.

    :param filename: the name of the file to save to
    :param packet_list: any iterable of objects of type ``Packet`` or
        ``TimestampPacket``.
    :param mode: optional, the "file mode" to open the data file
        (default: ``'a'``)
    :param version: optional, the LArPix+HDF5 format version to use. If
        writing a new file and version is unspecified or ``None``,
        the latest version will be used. If writing an existing file
        and version is unspecified or ``None``, the existing file's
        version will be used. If writing an existing file and version
        is specified and does not exactly match the existing file's
        version, a ``RuntimeError`` will be raised. (default: ``None``)

    '''
    with h5py.File(filename, mode) as f:
        # Create header
        if '_header' not in f.keys():
            header = f.create_group('_header')
            if version is None:
                version = latest_version
            header.attrs['version'] = version
            header.attrs['created'] = time.time()
        else:
            header = f['_header']
            file_version = header.attrs['version']
            if version is None:
                version = file_version
            elif header.attrs['version'] != version:
                raise RuntimeError('Incompatible versions: existing: %s, '
                    'specified: %s' % (file_version, version))
        header.attrs['modified'] = time.time()

        # Create datasets
        if version == '0.0':
            packet_dset_name = 'raw_packet'
        else:
            packet_dset_name = 'packets'
            direction_index = (
                    dtype_property_index_lookup[version]['packets']
                    ['direction'])
            if version[0] == '2':
                packet_type_index = (
                        dtype_property_index_lookup[version]['packets']
                        ['packet_type'])
                fifo_diagnostics_enabled_index = (
                        dtype_property_index_lookup[version]['packets']
                        ['fifo_diagnostics_enabled'])
                trigger_type_index = (
                        dtype_property_index_lookup[version]['packets']
                        ['trigger_type'])
                dataword_index = (
                        dtype_property_index_lookup[version]['packets']
                        ['dataword'])
        packet_dtype = dtypes[version][packet_dset_name]
        if packet_dset_name not in f.keys():
            packet_dset = f.create_dataset(packet_dset_name, shape=(len(packet_list),),
                    maxshape=(None,), dtype=packet_dtype)
            if version[0] == '1' or version[0] == '2':
                if version[-1] == '2' and version[0] == '2':
                    packet_dset.attrs['packet_types'] = '''
0: 'data',
1: 'test',
2: 'config write',
3: 'config read',
4: 'timestamp',
5: 'message',
6: 'sync',
7: 'trigger,
'''
                else:
                    packet_dset.attrs['packet_types'] = '''
0: 'data',
1: 'test',
2: 'config write',
3: 'config read',
4: 'timestamp',
5: 'message',
'''
            start_index = 0
        else:
            packet_dset = f[packet_dset_name]
            start_index = packet_dset.shape[0]
            packet_dset.resize(start_index + len(packet_list), axis=0)
        if version != '0.0':
            message_dset_name = 'messages'
            message_dtype = dtypes[version][message_dset_name]
            if message_dset_name not in f.keys():
                message_dset = f.create_dataset(message_dset_name,
                        shape=(0,), maxshape=(None,),
                        dtype=message_dtype)
                message_start_index = 0
            else:
                message_dset = f[message_dset_name]
                message_start_index = message_dset.shape[0]

        # Fill dataset
        encoded_packets = []
        messages = []
        for i, packet in enumerate(packet_list):
            if packet.__class__ in _format_method_lookup[version].get(packet_dset_name, tuple()):
                encoded_packet = _format_method_lookup[version][packet_dset_name][packet.__class__](packet)
                for idx in range(len(encoded_packet)):
                    if encoded_packet[idx] is None:
                        encoded_packet[idx] = 0
                encoded_packets.append(tuple(encoded_packet))

            if version != '0.0' and packet.__class__ in _format_method_lookup[version].get(message_dset_name, tuple()):
                encoded_message = _format_method_lookup[version][message_dset_name][packet.__class__](packet, counter=message_start_index + len(messages))
                messages.append(encoded_message)

        packet_dset[start_index:] = encoded_packets
        if version != '0.0':
            message_dset.resize(message_start_index + len(messages), axis=0)
            message_dset[message_start_index:] = messages

def from_file(filename, version=None, start=None, end=None):
    '''
    Read the data from the given file into LArPix Packet objects.

    :param filename: the name of the file to read
    :param version: the format version. Specify this parameter to
        enforce a version check. When a specific version such as
        ``'1.5'`` is specified, a ``RuntimeError`` will be raised if the
        stored format version number is not an exact match. If a version
        is prefixed with ``'~'`` such as ``'~1.5'``, a ``RuntimeError``
        will be raised if the stored format version is *incompatible*
        with the specified version. Compatible versions are those with
        the same major version and at least the same minor version. E.g.
        for ``'~1.5'``, versions between v1.5 and v2.0 are compatible.
        If unspecified or ``None``, will use the stored format version.
    :param start: the index of the first row to read
    :param end: the index after the last row to read (same semantics as
        Python ``range``)
    :returns packet_dict: a dict with keys ``'packets'`` containing a
        list of packet objects; and ``'created'``, ``'modified'``, and
        ``'version'``, containing the file metadata.

    '''
    with h5py.File(filename, 'r') as f:
        file_version = f['_header'].attrs['version']
        if version is None:
            version = file_version
        elif version[0] == '~':
            file_major, _, file_minor = file_version.split('.')
            version_major, _, version_minor = version.split('.')
            version_major = version_major[1:]
            if (file_major != version_major
                    or file_minor < version_minor):
                raise RuntimeError('Incompatible versions: existing: %s, '
                    'specified: %s' % (file_version, version))
            else:
                version = file_version
        elif version == file_version:
            pass
        else:
            raise RuntimeError('Incompatible versions: existing: %s, '
                'specified: %s' % (file_version, version))

        if version not in dtypes:
            raise RuntimeError('Unknown version: %s' % version)

        if version == '0.0':
            dset_name = 'raw_packet'
            message_dset = None
        else:
            dset_name = 'packets'
            message_dset_name = 'messages'
            message_props = (
                    dtype_property_index_lookup[version][message_dset_name])
            message_dset = f[message_dset_name]
        props = dtype_property_index_lookup[version][dset_name]
        packets = []
        if start is None and end is None:
            dset_iter = f[dset_name]
        else:
            dset_iter = f[dset_name][start:end]
        for row in dset_iter:
            pkt = _parse_method_lookup[version][dset_name](row, message_dset)
            if pkt is not None:
                packets.append(pkt)
        return {
                'packets': packets,
                'created': f['_header'].attrs['created'],
                'modified': f['_header'].attrs['modified'],
                'version': f['_header'].attrs['version'],
                }

